
# Drop Dead: Digging Around
# (C) Copyright 2021 by Ville Helin and Jyri Luukkonen
# For more Drop Dead games visit http://www.ttursas.com and
# https://vhelin.itch.io/drop-dead-collection

import pygame
import random
import copy

class TileIds:
    EMPTY  = 0
    WALL   = 1
    SAND   = 2
    ORANGE = 3
    WORM   = 4
    ENEMY  = 5
    ROCK   = 6
    DOOR   = 7

class Direction:
    LEFT = 0
    UP = 1
    RIGHT = 2
    DOWN = 3

class ScoreLabel:
    def __init__(self, px:int, py:int, text:str):
        self.px = px
        self.py = py
        self.text = text
        self.alive = 60

class GameObject:
    def __init__(self, tile_id:int):
        self.is_still_counter = 4
        self.processed_frame_id = -1
        if tile_id == TileIds.ENEMY:
            # enemy
            self.direction = random.randint(0, 3)

    def can_be_eaten(self) -> bool:
        return self.is_still_counter > 0

    def moved_by_worm(self, px:int, py:int, game:"Game"):
        # if the worm moves us we are ready to drop the next run() call
        # if there is nothing below us
        if game.tilemap[py+1][px] == TileIds.EMPTY:
            self.is_still_counter = 0

    def kill_worm_if_next_to_me(self, px:int, py:int, game:"Game"):
        if game.tilemap[py+1][px] == TileIds.WORM:
            game.worm_died()
        if game.tilemap[py-1][px] == TileIds.WORM:
            game.worm_died()
        if game.tilemap[py][px+1] == TileIds.WORM:
            game.worm_died()
        if game.tilemap[py][px-1] == TileIds.WORM:
            game.worm_died()

    def try_to_turn_left(self, px:int, py:int, game:"Game") -> int:
        move_x = 0
        move_y = 1
        try_direction = Direction.DOWN
        if self.direction == Direction.UP:
            move_x = -1
            move_y = 0
            try_direction = Direction.LEFT
        elif self.direction == Direction.RIGHT:
            move_x = 0
            move_y = -1
            try_direction = Direction.UP
        elif self.direction == Direction.DOWN:
            move_x = 1
            move_y = 0
            try_direction = Direction.RIGHT

        tx = px + move_x
        ty = py + move_y

        # cannot turn into tiles on the border
        if tx <= 0 or tx >= game.tiles_x-1 or ty <= 0 or ty >= game.tiles_y-1:
            return -1

        tile_id = game.tilemap[ty][tx]
        if tile_id == TileIds.EMPTY:
            # can turn left
            return try_direction
        else:
            # cannot turn left
            return -1

    def move(self, px:int, py:int, nx:int, ny:int, game:"Game"):
        tile_id = game.tilemap[py][px]
        game_object = game.game_objects[py][px]
        # clear the old position
        game.tilemap[py][px] = TileIds.EMPTY
        game.game_objects[py][px] = None
        # move to the new position
        game.tilemap[ny][nx] = tile_id
        game.game_objects[ny][nx] = game_object

    def run_enemy_ai(self, px:int, py:int, game:"Game"):
        self.kill_worm_if_next_to_me(px, py, game)

        # try to turn left
        left_turn_was_blocked = self.try_to_turn_left(px, py, game) < 0

        # try to move forward
        move_x = -1
        move_y = 0
        if self.direction == Direction.UP:
            move_x = 0
            move_y = -1
        elif self.direction == Direction.RIGHT:
            move_x = 1
            move_y = 0
        elif self.direction == Direction.DOWN:
            move_x = 0
            move_y = 1

        tx = px + move_x
        ty = py + move_y

        tile_id = game.tilemap[ty][tx]
        can_move = False
        if tx == 0 or tx == game.tiles_x-1 or ty == 0 or ty == game.tiles_y-1:
            # don't go into border tiles -> turn 90 degrees right
            self.direction = (self.direction + 1) % 4
        elif tile_id == TileIds.EMPTY:
            # empty
            can_move = True
        elif tile_id == TileIds.WORM:
            # worm
            can_move = True
            game.worm_died()
        else:
            # could not move -> turn 90 degrees right
            self.direction = (self.direction + 1) % 4

        if can_move:
            self.move(px, py, tx, ty, game)
            self.kill_worm_if_next_to_me(tx, ty, game)

        if can_move and left_turn_was_blocked:
            # try to turn left again
            left_direction = self.try_to_turn_left(tx, ty, game)
            if left_direction >= 0:
                # now can turn left!
                self.direction = left_direction

    def run(self, px:int, py:int, game:"Game", frame_id:int):
        # process only once a frame
        if self.processed_frame_id == frame_id:
            return
        self.processed_frame_id = frame_id

        tile_id = game.tilemap[py][px]
        if tile_id == TileIds.ENEMY:
            # enemies
            self.run_enemy_ai(px, py, game)
        elif tile_id == TileIds.ORANGE or tile_id == TileIds.ROCK:
            # oranges and rocks fall down unless on the 2nd bottommost row
            if py >= game.tiles_y-2:
                self.is_still_counter = 4
                return
            nx = px
            ny = py
            tile_id_down = game.tilemap[py+1][px]
            if tile_id_down == TileIds.EMPTY:
                self.is_still_counter -= 1
                ny += 1
            elif tile_id_down == TileIds.ORANGE or tile_id_down == TileIds.ROCK:
                # when on top of another orange or rock the object slips one tile sideways
                # if there is nothing below there...
                if game.tilemap[py][px-1] == TileIds.EMPTY and game.tilemap[py+1][px-1] == TileIds.EMPTY:
                    self.is_still_counter -= 1
                    nx -= 1
                elif game.tilemap[py][px+1] == TileIds.EMPTY and game.tilemap[py+1][px+1] == TileIds.EMPTY:
                    self.is_still_counter -= 1
                    nx += 1
                else:
                    self.is_still_counter = 4
            else:
                killed = False
                if self.is_still_counter <= 0:
                    if tile_id_down == TileIds.WORM:
                        # kill the player
                        game.worm_died()
                        killed = True
                        ny += 1
                    if tile_id_down == TileIds.ENEMY:
                        # kill enemies
                        game.score += 500
                        killed = True
                        ny += 1
                        game.add_score_label(nx, ny, str(500))
                if not killed:
                    self.is_still_counter = 4
            
            if self.is_still_counter <= 0:
                self.move(px, py, nx, ny, game)
                if ny == py+1 and nx == px:
                    # if we fell straight down, make oranges and rocks on top of us to fell right away, too
                    tile_id = game.tilemap[py-1][nx]
                    if tile_id == TileIds.ORANGE or tile_id == TileIds.ROCK:
                        game.game_objects[py-1][nx].is_still_counter = 0

class Game:
    def __init__(self):
        pygame.init()

        # here we decide if we use code generated graphics or bitmap sprites   
        self.use_sprites = True

        self.load_images()
        self.create_title_screen()
        self.screen_id = "Title"
        self.font_big = pygame.font.SysFont("Arial", 75, True)
        self.font_normal = pygame.font.SysFont("Arial", 20, True)
        self.font_small = pygame.font.SysFont("Arial", 12, True)
        self.clock = pygame.time.Clock()

        self.tiles_y = len(self.tilemap)
        self.tiles_x = len(self.tilemap[0])
        self.tilesize = 32
        self.top_score = 0
        self.frame_id = 0
        self.worm_move_x = 0
        self.worm_move_y = 0
        self.worm_x = -1
        self.worm_y = -1
        self.worm_direction = 0
        self.worm_direction_pressed = 0
        self.game_objects = []
        self.worm_is_dead = False
        self.level_completed = False

        self.screen_height = self.tilesize * self.tiles_y
        self.screen_width = self.tilesize * self.tiles_x
        self.screen = pygame.display.set_mode((self.screen_width, self.screen_height))

        pygame.display.set_caption("Drop Dead: Digging Around")

        self.main_loop()

    def load_images(self):
        if self.use_sprites == False:
            return

        self.images = []
        for name in ["0-empty", "1-wall", "2-sand", "3-orange", "worm-left", "5-enemy", "6-rock", "7-door"]:
            self.images.append(pygame.image.load("src/sprites/" + name + ".png"))
        self.image_worm_left = pygame.image.load("src/sprites/worm-left.png")
        self.image_worm_right = pygame.image.load("src/sprites/worm-right.png")
        self.image_worm_up = pygame.image.load("src/sprites/worm-up.png")
        self.image_worm_down = pygame.image.load("src/sprites/worm-down.png")
        self.image_title = pygame.image.load("src/sprites/" + "sprite-title.png")

    def create_title_screen(self):
        self.worm_x = -1
        self.worm_y = -1
        self.worm_direction = 0
        self.score_labels = []
        self.tilemap = [[1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                        [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                        [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                        [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                        [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                        [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                        [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                        [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                        [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                        [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                        [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                        [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                        [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                        [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                        [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]]

    def new_game(self):
        self.lives = 3
        self.score = 0
        self.level = 0
        self.create_level()

    def create_level(self):
        self.worm_direction = 0
        self.score_labels = []

        iterations = 0

        while True:
            self.tilemap = [[1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                            [1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1],
                            [1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1],
                            [1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1],
                            [1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1],
                            [1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1],
                            [1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1],
                            [1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1],
                            [1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1],
                            [1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1],
                            [1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1],
                            [1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1],
                            [1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1],
                            [1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1],
                            [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]]

            added = 0

            # add enemies
            enemies = int(self.level / 2)
            for i in range(0, enemies):
                for j in range(0, 256):
                    tx = random.randint(0, self.tiles_x-1)
                    ty = random.randint(0, self.tiles_y-1)
                    if self.is_on_borders(tx, ty):
                        continue
                    cx = random.randint(-1, 1) + tx
                    cy = random.randint(-1, 1) + ty
                    if self.tilemap[cy][cx] != TileIds.EMPTY and self.tilemap[cy][cx] != TileIds.SAND:
                        continue
                    # success!
                    added += 1
                    # clear the surrounding area
                    for py in range(ty-1, ty+2):
                        for px in range(tx-1, tx+2):
                            if self.tilemap[py][px] == TileIds.SAND:
                                self.tilemap[py][px] = TileIds.EMPTY
                    # place the enemy
                    self.tilemap[cy][cx] = TileIds.ENEMY
                    break

            # add rocks
            rocks = int((self.level + 1) * 1.5)
            for i in range(0, rocks):
                for j in range(0, 256):
                    tx = random.randint(0, self.tiles_x-1)
                    ty = random.randint(0, self.tiles_y-1)
                    if self.is_on_borders(tx, ty):
                        continue
                    if self.tilemap[ty][tx] != TileIds.EMPTY and self.tilemap[ty][tx] != TileIds.SAND:
                        continue
                    if self.tilemap[ty+1][tx] != TileIds.WALL and self.tilemap[ty+1][tx] != TileIds.SAND and self.tilemap[ty+1][tx] != TileIds.ROCK:
                        continue
                    # success!
                    added += 1
                    # place the rock
                    self.tilemap[ty][tx] = TileIds.ROCK
                    break

            # add oranges
            oranges = self.level + 1
            for i in range(0, oranges):
                for j in range(0, 256):
                    tx = random.randint(0, self.tiles_x-1)
                    ty = random.randint(0, self.tiles_y-1)
                    if self.is_on_borders(tx, ty):
                        continue
                    if self.tilemap[ty][tx] != TileIds.EMPTY and self.tilemap[ty][tx] != TileIds.SAND:
                        continue
                    if self.tilemap[ty+1][tx] != TileIds.WALL and self.tilemap[ty+1][tx] != TileIds.SAND and self.tilemap[ty+1][tx] != TileIds.ROCK and self.tilemap[ty+1][tx] != TileIds.ORANGE:
                        continue
                    # success!
                    added += 1
                    # place the orange
                    self.tilemap[ty][tx] = TileIds.ORANGE
                    break

            # add worm
            worms = 1
            for j in range(0, 256):
                tx = random.randint(0, self.tiles_x-1)
                ty = random.randint(0, self.tiles_y-1)
                if self.is_on_borders(tx, ty):
                    continue
                if self.tilemap[ty][tx] != TileIds.EMPTY and self.tilemap[ty][tx] != TileIds.SAND:
                    continue
                if self.tilemap[ty+1][tx] != TileIds.WALL and self.tilemap[ty+1][tx] != TileIds.SAND and self.tilemap[ty+1][tx] != TileIds.ROCK and self.tilemap[ty+1][tx] != TileIds.ORANGE:
                    continue
                # check the surrounding area
                ok = True
                for py in range(ty-1, ty+2):
                    for px in range(tx-1, tx+2):
                        if self.tilemap[py][px] == TileIds.EMPTY or self.tilemap[py][px] == TileIds.ENEMY:
                            ok = False
                if ok == False:
                    continue
                # success!
                added += 1
                # place the worm
                self.tilemap[ty][tx] = TileIds.WORM
                break

            # did we manage to add all the items?
            if added < enemies + rocks + oranges + worms:
                # no...
                iterations += 1
                if iterations >= 1024:
                    # give up and use the last generated level
                    break
            else:
                # all done successfully
                break
        
        self.worm_is_dead = False
        self.level_completed = False
        self.worm_move_x = 0
        self.worm_move_y = 0

        # remember the level for self.replay_level()
        self.tilemap_saved = copy.deepcopy(self.tilemap)
        self.process_tilemap()

    def replay_level(self):
        self.worm_is_dead = False
        self.level_completed = False
        self.tilemap = copy.deepcopy(self.tilemap_saved)
        self.process_tilemap()

    def process_tilemap(self):
        # find the worm
        for y in range(0, self.tiles_y):
            for x in range(0, self.tiles_x):
                if self.tilemap[y][x] == TileIds.WORM:
                    self.worm_x = x
                    self.worm_y = y

        # find rocks, oranges and enemies and turn them into GameObjects
        self.game_objects = []
        for y in range(0, self.tiles_y):
            self.game_objects.append([])
            for x in range(0, self.tiles_x):
                self.game_objects[y].append(None)
                tile_id = self.tilemap[y][x]
                if tile_id == TileIds.ORANGE or tile_id == TileIds.ENEMY or tile_id == TileIds.ROCK:
                    self.game_objects[y][x] = GameObject(tile_id)

    def is_on_borders(self, tx:int, ty:int) -> bool:
        if tx <= 0 or tx >= self.tiles_x-1 or ty <= 0 or ty >= self.tiles_y-1:
            return True
        else:
            return False

    def main_loop(self):
        while True:
            self.screen.fill((0, 0, 0))

            if self.screen_id == "Title":
                self.process_events_title()
                self.draw_graphics_title()
            elif self.screen_id == "InGame":
                self.process_events_ingame()
                self.draw_graphics_ingame()

                # process ingame logic on special frames
                if self.frame_id % 10 == 0:
                    self.run_game()

                # remember the top score
                if self.score > self.top_score:
                    self.top_score = self.score

            pygame.display.flip()
            self.clock.tick(60)

            self.frame_id += 1

    def worm_died(self):
        # clear the worm's tile
        self.tilemap[self.worm_y][self.worm_x] = TileIds.EMPTY

        # reset level completition status
        self.level_completed = False
        self.level_completed_timer = 0

        # mark him dead
        self.worm_is_dead = True
        self.worm_is_dead_timer = 0
        self.worm_move_x = 0
        self.worm_move_y = 0

        # add RIP
        self.add_score_label(self.worm_x, self.worm_y, "RIP")

    def process_common_event(self, e):
        if e.type == pygame.QUIT:
            exit()
        if e.type == pygame.KEYDOWN:
            if e.key == pygame.K_ESCAPE:
                if self.screen_id == "InGame":
                    self.create_title_screen()
                    self.screen_id = "Title"
                else:
                    exit()

    def process_events_title(self):
        for e in pygame.event.get():
            self.process_common_event(e)
            if e.type == pygame.KEYDOWN:
                if e.key == pygame.K_SPACE:
                    self.screen_id = "InGame"
                    self.new_game()

    def process_events_ingame(self):
        for e in pygame.event.get():
            self.process_common_event(e)
            if not self.worm_is_dead:
                if e.type == pygame.KEYDOWN:
                    # cursor keys
                    if e.key == pygame.K_LEFT:
                        self.worm_direction_pressed = 0
                        self.worm_move_x = -1
                        self.worm_move_y = 0
                    elif e.key == pygame.K_RIGHT:
                        self.worm_direction_pressed = 2
                        self.worm_move_x = 1
                        self.worm_move_y = 0
                    elif e.key == pygame.K_UP:
                        self.worm_direction_pressed = 1
                        self.worm_move_y = -1
                        self.worm_move_x = 0
                    elif e.key == pygame.K_DOWN:
                        self.worm_direction_pressed = 3
                        self.worm_move_y = 1
                        self.worm_move_x = 0
                    # aswd
                    if e.key == pygame.K_a:
                        self.worm_direction_pressed = 0
                        self.worm_move_x = -1
                        self.worm_move_y = 0
                    elif e.key == pygame.K_d:
                        self.worm_direction_pressed = 2
                        self.worm_move_x = 1
                        self.worm_move_y = 0
                    elif e.key == pygame.K_w:
                        self.worm_direction_pressed = 1
                        self.worm_move_y = -1
                        self.worm_move_x = 0
                    elif e.key == pygame.K_s:
                        self.worm_direction_pressed = 3
                        self.worm_move_y = 1
                        self.worm_move_x = 0
                if e.type == pygame.KEYUP:
                    # cursor keys
                    if e.key == pygame.K_LEFT:
                        self.worm_move_x = 0
                    elif e.key == pygame.K_RIGHT:
                        self.worm_move_x = 0
                    elif e.key == pygame.K_UP:
                        self.worm_move_y = 0
                    elif e.key == pygame.K_DOWN:
                        self.worm_move_y = 0
                    # aswd
                    if e.key == pygame.K_a:
                        self.worm_move_x = 0
                    elif e.key == pygame.K_d:
                        self.worm_move_x = 0
                    elif e.key == pygame.K_w:
                        self.worm_move_y = 0
                    elif e.key == pygame.K_s:
                        self.worm_move_y = 0

    def run_game(self):
        self.move_worm(self.worm_move_x, self.worm_move_y, self.worm_direction_pressed)

        # the player is dead?
        if self.worm_is_dead:
            self.worm_is_dead_timer += 1
            if self.worm_is_dead_timer == 15:
                self.lives -= 1
                if self.lives > 0:
                    self.replay_level()
                else:
                    pass
            elif self.worm_is_dead_timer == 30:
                self.worm_is_dead = False
                self.create_title_screen()
                self.screen_id = "Title"

        # did the player complete the level?
        if self.level_completed:
            self.level_completed_timer += 1
            if self.level_completed_timer == 6:
                self.level += 1
                self.create_level()

        # run the GameObjects from bottom to up as gravity goes down
        for y in range(0, self.tiles_y):
            py = self.tiles_y-1 - y
            for x in range(0, self.tiles_x):
                if self.game_objects[py][x] != None:
                    self.game_objects[py][x].run(x, py, self, self.frame_id)

    def move_worm(self, mx:int, my:int, direction:int):
        # no movement?
        if mx == 0 and my == 0:
            return
        # has the level ended?
        if self.level_completed:
            return

        self.worm_direction = direction

        tx = self.worm_x + mx
        if tx < 0 or tx >= self.tiles_x:
            return
        ty = self.worm_y + my
        if ty < 0 or ty >= self.tiles_y:
            return

        can_move = False
        tile_id = self.tilemap[ty][tx]
        if tile_id == TileIds.EMPTY or tile_id == TileIds.SAND:
            can_move = True
        elif tile_id == TileIds.ORANGE:
            if self.game_objects[ty][tx].can_be_eaten():
                self.score += 100
                self.add_score_label(tx, ty, str(100))
                can_move = True
        elif tile_id == TileIds.ENEMY:
            self.worm_died()
            # clear the old tile
            self.tilemap[self.worm_y][self.worm_x] = TileIds.EMPTY
        elif tile_id == TileIds.ROCK:
            if mx != 0:
                # is there room behind the rock?
                nx = tx + mx
                if self.tilemap[ty][nx] == TileIds.EMPTY:
                    # yes -> move the rock
                    self.tilemap[ty][nx] = self.tilemap[ty][tx]
                    self.tilemap[ty][tx] = TileIds.EMPTY
                    self.game_objects[ty][nx] = self.game_objects[ty][tx]
                    self.game_objects[ty][nx].moved_by_worm(nx, ty, self)
                    self.game_objects[ty][tx] = None
                    can_move = True
        elif tile_id == TileIds.DOOR:
            can_move = True

        if can_move:
            # clear the old tile
            self.tilemap[self.worm_y][self.worm_x] = TileIds.EMPTY

            # enter the new tile
            self.tilemap[ty][tx] = TileIds.WORM
            self.game_objects[ty][tx] = None

            self.worm_x = tx
            self.worm_y = ty

            if tile_id == TileIds.ORANGE:
                self.out_of_oranges()
            if tile_id == TileIds.DOOR:
                self.level_completed = True
                self.level_completed_timer = 0

    def is_suitable_tile_for_door(self, x:int, y:int) -> bool:
        # only wall tiles are accepted
        if self.tilemap[y][x] != TileIds.WALL:
            return False
        # cannot accept bottom corners
        if (x == 0 and y == self.tiles_y-1) or (x == self.tiles_x-1 and y == self.tiles_y-1):
            return False
        # is there a rock blocking the door?
        if x == 0 and self.tilemap[y][x+1] == TileIds.ROCK:
            return False
        if x == self.tiles_x-1 and self.tilemap[y][x-1] == TileIds.ROCK:
            return False
        if y == self.tiles_y-1 and self.tilemap[y-1][x] == TileIds.ROCK:
            return False

        return True

    def add_score_label(self, px:int, py:int, text:str):
        self.score_labels.append(ScoreLabel(px*self.tilesize + self.tilesize/2, py*self.tilesize + self.tilesize/2, text))

    def out_of_oranges(self):
        # any oranges left?
        for y in range(0, self.tiles_y):
            for x in range(0, self.tiles_x):
                if self.tilemap[y][x] == TileIds.ORANGE:
                    return

        # no oranges found, do we have a door?
        for y in range(0, self.tiles_y):
            for x in range(0, self.tiles_x):
                if self.tilemap[y][x] == TileIds.DOOR:
                    return

        # no doors found, so lets open one (no doors at the top)
        candidates = 0
        for y in range(1, self.tiles_y):
            for x in range(0, self.tiles_x):
                if self.is_suitable_tile_for_door(x, y):
                    candidates += 1

        id = random.randint(0, candidates-1)

        i = 0
        for y in range(1, self.tiles_y):
            for x in range(0, self.tiles_x):
                if self.is_suitable_tile_for_door(x, y):
                    if i == id:
                        self.tilemap[y][x] = TileIds.DOOR
                    i += 1

    def draw_tile(self, tile_id:int, coordinates:tuple):
        if self.use_sprites:
            if tile_id == TileIds.WORM:
                if self.worm_direction == Direction.LEFT:
                    # left
                    self.screen.blit(self.image_worm_left, coordinates)
                elif self.worm_direction == 1:
                    # up
                    self.screen.blit(self.image_worm_up, coordinates)
                elif self.worm_direction == 2:
                    # right
                    coordinates = (coordinates[0] - 8, coordinates[1])
                    self.screen.blit(self.image_worm_right, coordinates)
                elif self.worm_direction == 3:
                    # down
                    coordinates = (coordinates[0], coordinates[1] - 8)
                    self.screen.blit(self.image_worm_down, coordinates)
            else:
                self.screen.blit(self.images[tile_id], coordinates)
        else:
            if tile_id == TileIds.EMPTY:
                # 0 - empty
                pygame.draw.rect(self.screen, (0, 0, 0), pygame.Rect(coordinates[0], coordinates[1], self.tilesize, self.tilesize))
            elif tile_id == TileIds.WALL:
                # 1 - wall
                pygame.draw.rect(self.screen, (136, 112, 0), pygame.Rect(coordinates[0], coordinates[1], self.tilesize, self.tilesize))
            elif tile_id == TileIds.SAND:
                # 2 - sand
                pygame.draw.rect(self.screen, (240, 188, 60), pygame.Rect(coordinates[0], coordinates[1], self.tilesize, self.tilesize))
            elif tile_id == TileIds.ORANGE:
                # 3 - orange
                px = coordinates[0] + self.tilesize/2
                py = coordinates[1] + self.tilesize/2
                coordinates = (px, py)

                pygame.draw.circle(self.screen, (244, 104, 7), coordinates, self.tilesize/2-1)
                pygame.draw.circle(self.screen, (139, 143, 66), (coordinates[0], coordinates[1] - self.tilesize/4), self.tilesize/8)
            elif tile_id == TileIds.WORM:
                # 4 - worm
                px = coordinates[0] + self.tilesize/2
                py = coordinates[1] + self.tilesize/2
                coordinates = (px, py)

                cx = coordinates[0]
                cy = coordinates[1]
                ims = self.tilesize
                i4 = ims / 4

                if self.worm_direction == Direction.LEFT:
                    # left
                    pygame.draw.circle(self.screen, (72, 219, 68), coordinates, self.tilesize/2.25-1)
                    pygame.draw.circle(self.screen, (72, 219, 68), (cx + i4*1.5, cy + i4*1.0), self.tilesize/5)
                    pygame.draw.circle(self.screen, (72, 219, 68), (cx + i4*2.25, cy + i4*1.4), self.tilesize/7)
                elif self.worm_direction == Direction.UP:
                    # up
                    pygame.draw.circle(self.screen, (72, 219, 68), coordinates, self.tilesize/2.25-1)
                    pygame.draw.circle(self.screen, (72, 219, 68), (cx, cy + i4*1.5), self.tilesize/5)
                    pygame.draw.circle(self.screen, (72, 219, 68), (cx, cy + i4*2.25), self.tilesize/7)
                elif self.worm_direction == Direction.RIGHT:
                    # right
                    pygame.draw.circle(self.screen, (72, 219, 68), coordinates, self.tilesize/2.25-1)
                    pygame.draw.circle(self.screen, (72, 219, 68), (cx - i4*1.5, cy + i4*1.0), self.tilesize/5)
                    pygame.draw.circle(self.screen, (72, 219, 68), (cx - i4*2.25, cy + i4*1.4), self.tilesize/7)
                elif self.worm_direction == Direction.DOWN:
                    # down
                    pygame.draw.circle(self.screen, (72, 219, 68), coordinates, self.tilesize/2.25-1)
                    pygame.draw.circle(self.screen, (72, 219, 68), (cx, cy - i4*1.5), self.tilesize/5)
                    pygame.draw.circle(self.screen, (72, 219, 68), (cx, cy - i4*2.25), self.tilesize/7)

                pygame.draw.circle(self.screen, (255, 255, 255), (cx - i4, cy), i4)
                pygame.draw.circle(self.screen, (0, 0, 0), (cx - i4, cy), i4/2)

                pygame.draw.circle(self.screen, (255, 255, 255), (cx + i4, cy), i4)
                pygame.draw.circle(self.screen, (0, 0, 0), (cx + i4, cy), i4/2)

                pygame.draw.circle(self.screen, (0, 0, 0), (cx, cy + i4), i4/3)
            elif tile_id == TileIds.ENEMY:
                # 5 - enemy
                vertices = [(16,0), (27, 3), (30, 6), (30, 31), (25, 25), (20, 31), (15, 25), (10, 31), (5, 25), (1, 31), (1, 6), (4, 3)]
                for i in range(0, len(vertices)):
                    px = coordinates[0] + vertices[i][0]
                    py = coordinates[1] + vertices[i][1]
                    vertices[i] = (px, py)
                pygame.draw.polygon(self.screen, (255, 255, 255), vertices, 0)

                px = coordinates[0] + 9
                py = coordinates[1] + 13
                pygame.draw.circle(self.screen, (0, 0, 0), (px, py), self.tilesize/6)
                px = coordinates[0] + 9 + 14
                py = coordinates[1] + 13
                pygame.draw.circle(self.screen, (0, 0, 0), (px, py), self.tilesize/6)

                pygame.draw.rect(self.screen, (0, 0, 0), pygame.Rect(coordinates[0] + self.tilesize/2 - 10, coordinates[1] + self.tilesize/2 + 5, 20, 2))
            elif tile_id == TileIds.ROCK:
                # 6 - rock
                px = coordinates[0] + self.tilesize/2
                py = coordinates[1] + self.tilesize/2
                coordinates = (px, py)

                pygame.draw.circle(self.screen, (102, 64, 47), coordinates, self.tilesize/2-1)
            elif tile_id == TileIds.DOOR:
                # 7 - door
                pygame.draw.rect(self.screen, (136, 112, 0), pygame.Rect(coordinates[0], coordinates[1], self.tilesize, self.tilesize))
                pygame.draw.rect(self.screen, (0, 0, 0), pygame.Rect(coordinates[0]+2, coordinates[1]+2, self.tilesize-4, self.tilesize-4))

    def draw_blood_splatter(self, x:int, y:int):
        x += self.tilesize/2
        y += self.tilesize/2

        pygame.draw.circle(self.screen, (255, 0, 0), (x, y), self.tilesize/2-1)

    def draw_tilemap(self):
        # draw all tiles once
        for y in range(self.tiles_y):
            for x in range(self.tiles_x):
                tile_id = self.tilemap[y][x]
                drew_blood = False
                # the player is dead?
                if x == self.worm_x and y == self.worm_y and self.worm_is_dead:
                    # draw the blood splatter behind some objects
                    if tile_id == TileIds.ORANGE or tile_id == TileIds.ENEMY or tile_id == TileIds.ROCK:
                        self.draw_blood_splatter(self.worm_x*self.tilesize, self.worm_y*self.tilesize)
                        drew_blood = True
                self.draw_tile(tile_id, (x * self.tilesize, y * self.tilesize))
                # the player is dead?
                if x == self.worm_x and y == self.worm_y and self.worm_is_dead:
                    # draw the blood splatter on top of some objects
                    if drew_blood == False:
                        self.draw_blood_splatter(self.worm_x*self.tilesize, self.worm_y*self.tilesize)

        if not self.worm_is_dead:
            # redraw the worm as it spills out from its tile
            self.draw_tile(TileIds.WORM, (self.worm_x * self.tilesize, self.worm_y * self.tilesize))

    def draw_graphics_title(self):
        self.draw_tilemap()

        if self.use_sprites:
            self.screen.blit(self.image_title, (self.screen_width/2 - self.image_title.get_width()/2,
                                                self.screen_height/2 - self.image_title.get_height()/2))
        else:
            text = self.font_big.render("DROP DEAD", True, (213, 0, 40))
            self.screen.blit(text, (self.screen_width/2-text.get_width()/2, 75))
            text = self.font_normal.render("DIGGING AROUND", True, (255, 255, 255))
            self.screen.blit(text, (self.screen_width/2-text.get_width()/2, 150))

            text = self.font_normal.render("USE CURSOR KEYS TO MOVE LITTLE WORM", True, (255, 255, 255))
            self.screen.blit(text, (self.screen_width/2-text.get_width()/2-24, 200+20))
            text = self.font_normal.render("EAT ORANGES", True, (255, 255, 255))
            self.screen.blit(text, (self.screen_width/2-text.get_width()/2, 250))
            text = self.font_normal.render("AVOID TERRIFYING ENEMIES", True, (255, 255, 255))
            self.screen.blit(text, (self.screen_width/2-text.get_width()/2, 300-20))

            text = self.font_normal.render("PRESS SPACE TO START", True, (255, 255, 255))
            self.screen.blit(text, (self.screen_width/2-text.get_width()/2, 350))

            text = self.font_small.render("FOR MORE DROP DEAD GAMES VISIT HTTP://WWW.TTURSAS.COM", True, (255, 255, 255))
            self.screen.blit(text, (self.screen_width/2-text.get_width()/2, 408))

            text = self.font_small.render("COPYRIGHT (C) 2021 VILLE HELIN AND JYRI LUUKKONEN", True, (255, 255, 255))
            self.screen.blit(text, (self.screen_width/2-text.get_width()/2, 420))

            self.draw_tile(TileIds.WORM, (524, 214))
            self.draw_tile(TileIds.ORANGE, (401, 246))
            self.draw_tile(TileIds.ENEMY, (472, 276))

        self.draw_text_with_outlines(self.font_normal, "TOP = " + str(self.top_score), (255, 255, 255), self.screen_width/2, 3, "Center")
        self.draw_text_with_outlines(self.font_normal, "VERSION 1.0", (255, 255, 255), self.screen_width/2, self.screen_height - 3, "Center", "Bottom")

    def draw_graphics_ingame(self):
        self.draw_tilemap()

        self.draw_text_with_outlines(self.font_normal, "LIVES = " + str(self.lives), (255, 255, 255), 3, 3)
        self.draw_text_with_outlines(self.font_normal, "SCORE = " + str(self.score), (255, 255, 255), self.screen_width/2, 3, "Center")
        self.draw_text_with_outlines(self.font_normal, "LEVEL = " + str(self.level + 1), (255, 255, 255), self.screen_width-3, 3, "Right")

        if self.worm_is_dead and self.lives <= 0 and self.worm_is_dead_timer > 6:
            self.draw_text_with_outlines(self.font_normal, "GAME OVER", (255, 255, 255), self.screen_width/2, self.screen_height/2, "Center", "Center")

        for s in self.score_labels:
            s.alive -= 1
            self.draw_text_with_outlines(self.font_small, s.text, (255, 255, 255), s.px, s.py, "Center", "Center")

        # remove dead ScoreLabels
        self.score_labels = list(filter(lambda s: s.alive > 0, self.score_labels))

    def draw_text_with_outlines(self, font:pygame.font.Font, text:str, color:tuple, pos_x:int, pos_y:int, alignment_x:str = "Left", alignment_y:str = "Top"):
        # paint the text in black around the real, white text (painted later)
        t = font.render(text, True, (0, 0, 0))

        if alignment_x == "Left":
            # align left
            pass
        elif alignment_x == "Center":
            # align center
            pos_x -= t.get_width()/2
        elif alignment_x == "Right":
            # align right
            pos_x -= t.get_width()

        if alignment_y == "Top":
            # align top
            pass
        elif alignment_y == "Center":
            pos_y -= t.get_height()/2
        elif alignment_y == "Bottom":
            pos_y -= t.get_height()

        pos_x = int(pos_x)
        pos_y = int(pos_y)

        for y in range(pos_y-1, pos_y+2):
            for x in range(pos_x-1, pos_x+2):
                self.screen.blit(t, (x, y))

        # paint the real text itself
        t = font.render(text, True, color)
        self.screen.blit(t, (pos_x, pos_y))

if __name__ == "__main__":
    Game()
